/*#######################################################################################
AVR DCF77 Clock 

Copyright (C) 2005 Ulrich Radig

#######################################################################################*/

#include "clock.h"

#include <avr/io.h>
#include <avr/interrupt.h>

#include <string.h>

#define SYSCLK	20000000


//Die Uhrzeit seht in folgenden Variablen
volatile unsigned char ss   = 0;   //Globale Variable fr die Sekunden
volatile unsigned char mm   = 0;   //Globale Variable fr die Minuten
volatile unsigned char hh   = 0;   //Globale Variable fr die Stunden
volatile unsigned char day   = 0;   //Globale Variable fr den Tag
volatile unsigned char mon   = 0;   //Globale Variable fr den Monat
volatile unsigned int year   = 0;   //Globale Variable fr das Jahr

//Bitzhler fr RX Bit
volatile unsigned char rx_bit_counter = 0;

//64 Bit fr DCF77 bentigt werden 59 Bits
volatile unsigned long long dcf_rx_buffer = 0;

//Hilfs Sekunden Counter
volatile unsigned int h_ss = 0;

//Hilfs Variable fr Stundenwechsel
volatile unsigned int h_hh = 0;

char Debug[20];

void Debug_Out(int line, int data)
{
	sprintf(Debug, "%08u", data);
	lcd_setcursor(0,line);
	lcd_string(Debug);
}



// Rundarray fr gemessene Werte
int pulses[20];

// Position im Rundarray
int curr_position; 

// Array nutzbar fr Berechnung?
int array_completed = 0;

// Adaptiver Mittelwert
int pulse_threshold = 0;

void Set_Pulse(int pulse)
{
	pulses[curr_position] = pulse;
	if (curr_position == 19)
	{
		curr_position = 0;
		array_completed = 1;
	}
	else
		curr_position++;
}



void Calc_Treshold(void)
{
	int i;
	long medium;
	long medium_high;
	long medium_low;
	int num_values;
	
	if (array_completed == 1)
	{
// Mittelwert berechnen		
		medium = 0;
		for (i=0; i<20; i++) 
		{
			medium = medium + pulses[i];
		}
		medium = medium / 20;
		
// oberen Mittelwert berechnen
		num_values = 0;
		medium_high = 0;
		for (i=0; i<20; i++) 
		{	
			if (pulses[i] > medium)
			{
				medium_high = medium_high + pulses[i];
				num_values++;
			}
		}
		medium_high = medium_high / num_values;
		
// unteren Mittelwert berechnen
		num_values = 0;
		medium_low = 0;
		for (i=0; i<20; i++) 
		{	
			if (pulses[i] < medium)
			{
				medium_low = medium_low + pulses[i];
				num_values++;
			}
		}
		medium_low = medium_low / num_values;		
		
// Detektionsschwelle als Mittelwert der beiden berechneten Mittelwerte
		pulse_threshold = (medium_low + medium_high) / 2;
			
	}
}



//###########################################################################
// Adaptive Anpassung der Detektionsschwelle

//###########################################################################



//############################################################################
//Overflow Interrupt wird ausgelst bei 59Sekunde oder fehlenden DCF77 Signal 
SIGNAL (SIG_OVERFLOW1)
//############################################################################
{   
   struct  DCF77_Bits *rx_buffer;
   rx_buffer = (struct DCF77_Bits*)(char*)&dcf_rx_buffer;
   
      
   //Zurcksetzen des Timers
   TCNT1 = 65535 - (SYSCLK / 1024);
   //wurden alle 59 Bits empfangen und sind die Paritys richtig?
   if (rx_bit_counter == 59 && 
      flags.parity_P1 == rx_buffer->P1 && 
      flags.parity_P2 == rx_buffer->P2 &&
      flags.parity_P3 == rx_buffer->P3)
      //Alle 59Bits empfangen stellen der Uhr nach DCF77 Buffer
      {
      //Berechnung der Minuten BCD to HEX
      mm = rx_buffer->Min-((rx_buffer->Min/16)*6);
        
        if (mm != 0){mm--;}else{mm = 59; h_hh = 1;};
      
        //Berechnung der Stunden BCD to HEX
      hh = rx_buffer->Hour-((rx_buffer->Hour/16)*6);

      if (h_hh) {hh--;h_hh = 0;};

      //Berechnung des Tages BCD to HEX
      day= rx_buffer->Day-((rx_buffer->Day/16)*6); 
      //Berechnung des Monats BCD to HEX
      mon= rx_buffer->Month-((rx_buffer->Month/16)*6);
      //Berechnung des Jahres BCD to HEX
      year= 2000 + rx_buffer->Year-((rx_buffer->Year/16)*6);
      //Sekunden werden auf 0 zurckgesetzt
      ss = 59;
      flags.dcf_sync = 1;
	  
	  
		PORTD |= _BV(PD0);	
      }
   else
      //nicht alle 59Bits empfangen bzw kein DCF77 Signal Uhr luft 
      //manuell weiter
      {
      Add_one_Second();
      flags.dcf_sync = 0;
	  PORTD &= ~_BV(PD0);
	  
      }
   //zurcksetzen des RX Bit Counters
   rx_bit_counter = 0;
   //Lschen des Rx Buffers
   dcf_rx_buffer = 0;
 };

//############################################################################
//DCF77 Modul empfngt Trger 
SIGNAL (DCF77_INT)
//############################################################################
{
    if (PORTC & _BV(PC0))
		PORTC &= ~_BV(PC0); //heartbeat led off
	else
		PORTC |= _BV(PC0);
    
   //Auswertung der Pulseweite 
   if (INT0_CONTROL == INT0_RISING_EDGE)
   {
		flags.dcf_rx ^= 1;
//Sekunden-Hilfs Counter berechnen // SYSCLK defined in USART.H
		h_ss = h_ss + TCNT1 - (65535 - (SYSCLK / 1024));
//Zurcksetzen des Timers
		TCNT1 = 65535 - (SYSCLK / 1024);
//ist eine Secunde verstrichen?
		if (h_ss > (SYSCLK / 1024 / 100 * 90)) //90% von 1Sekunde
        {
 //Addiere +1 zu Sekunden
			Add_one_Second();
//Zurcksetzen des Hilfs Counters
			h_ss = 0;
        };
//Nchster Interrupt wird ausgelst bei abfallender Flanke
		INT0_CONTROL = INT0_FALLING_EDGE;
   }
   else
   {
//Auslesen der Pulsweite von ansteigender Flanke zu abfallender Flanke
		unsigned int pulse_wide = TCNT1;
//Zurcksetzen des Timers
		TCNT1 = 65535 - (SYSCLK / 1024);
	 
//Secunden Hilfs Counter berechnen
		h_ss = h_ss + pulse_wide - (65535 - (SYSCLK / 1024));
//Parity speichern
//beginn von Bereich P1/P2/P3
		if (rx_bit_counter ==  21 || rx_bit_counter ==  29 || rx_bit_counter ==  36) 
        {
			flags.parity_err = 0;
        };
//Speichern von P1
		if (rx_bit_counter ==  28) 
		{
			flags.parity_P1 = flags.parity_err;
		};
//Speichern von P2
		if (rx_bit_counter ==  35) 
		{
			flags.parity_P2 = flags.parity_err;
		};
//Speichern von P3
		if (rx_bit_counter ==  58) 
		{
			flags.parity_P3 = flags.parity_err;
		};
	  
//berprfen ob eine 0 oder eine 1 empfangen wurde
//0 = 100ms
//1 = 200ms
	  
		Set_Pulse(pulse_wide);
		if (array_completed == 1)
		{
			Calc_Treshold();
//			Debug_Out(1, pulse_wide);
//	  		Debug_Out(2, pulse_threshold);
//Abfrage grer als 150ms (15% von 1Sekund also 150ms)   
//  	   	if (pulse_wide > (65535 - (SYSCLK / 1024)/100*85))
			if (pulse_wide > pulse_threshold)
			{
//Schreiben einer 1 im dcf_rx_buffer an der Bitstelle rx_bit_counter
				dcf_rx_buffer = dcf_rx_buffer | ((unsigned long long) 1 << rx_bit_counter);
//Toggle Hilfs Parity
				flags.parity_err = flags.parity_err ^ 1;
				PORTC &= ~_BV(PC1);
				PORTD |= _BV(PD4);
			}
			else 
			{
				PORTC |= _BV(PC1);
				PORTD &= ~_BV(PD4);
			}
		}
//Nchster Interrupt wird ausgelst bei ansteigender Flanke
		INT0_CONTROL = INT0_RISING_EDGE;
//RX Bit Counter wird um 1 incrementiert
		rx_bit_counter++;
    }
 
};

//############################################################################
//Addiert 1 Sekunde
void Add_one_Second (void)
//############################################################################
{
   ss++;//Addiere +1 zu Sekunden
   if (ss == 60)
   {
      ss = 0;
      mm++;//Addiere +1 zu Minuten
      if (mm == 60)
      {
         mm = 0;
         hh++;//Addiere +1 zu Stunden
         if (hh == 24)
         {
            hh = 0;
         }
      }
   }
};

//############################################################################
//Diese Routine startet und inizialisiert den Timer
void Start_Clock (void)
//############################################################################
{
   //Interrupt DCF77 einschalten auf ansteigende Flanke
   DCF77_INT_ENABLE();
   INT0_CONTROL = INT0_RISING_EDGE;  // startet bei aufsteigender Flanke
      
   //Interrupt Overflow enable
   TIMSK1 |= (1 << TOIE1);
   //Setzen des Prescaler auf 1024 
   TCCR1B |= (1<<CS10 | 0<<CS11 | 1<<CS12); 
    
   TCNT1 = 65535 - (SYSCLK / 1024);
   return;
};



